import sklearn
import os
import numpy as np
from discovery_imaging_utils import dictionary_utils
import matplotlib.pyplot as plt

def calc_triple_network_model(parcellated_timeseries, parcel_ids):
    
    salience_ids = []
    control_ids = []
    dmn_ids = []

    salience_identifier = 'SalVentAttnA'
    executive_control_identifier = 'ContA'
    dmn_identifier = 'DefaultA'

    for i, temp_label in enumerate(parcel_ids):

        if salience_identifier in temp_label:

            salience_ids.append(i)

        if executive_control_identifier in temp_label:

            control_ids.append(i)

        if dmn_identifier in temp_label:

            dmn_ids.append(i)


    salience_signal = np.mean(parcellated_timeseries[salience_ids,:],axis=0)
    control_signal = np.mean(parcellated_timeseries[control_ids,:], axis=0)
    dmn_signal = np.mean(parcellated_timeseries[dmn_ids,:], axis=0)
    
    nii_mean, nii_std, nii_corr, z_corr_1, z_corr_2 = calc_network_interaction_index(control_signal, dmn_signal, salience_signal, 0.8)
    
    return nii_mean, nii_std, nii_corr
        
    

def calc_network_interaction_index(timeseries_1, timeseries_2, timeseries_a, TR, window_length_seconds=40, slide_step_seconds=2, decay_constant=0.333):
    
    import numpy as np
    
    #This function calculates the network interaction index as defined
    #by the paper Dysregulated Brain Dynamics in a Triple-Network Saliency 
    #Model of Schizophrenia and Its Relation to Psychosis published in Biological
    #Psychiatry by Supekar et. al., which follows from Time-Resolved Resting-State
    #Brain Networks published in PNAS by Zalesky et. al. 
    
    #For the triple salience model, timeseries_1 should be a cleaned central executive
    #network time signal, timeseries_2 should be a default mode time signal, and
    #timeseries_a should be the salience signal. TR must be defined. The window length
    #defaults to 40s, and slide step defaults to 2s. The decay constant for the window
    #weights defaults to 0.333, and higher values make the weighting approach linearity.
    
    #The window length and slide step will be rounded to the nearest TR. 
    
    #The function will output a nii_mean, nii_std, and nii_corr, where (without including
    #notation for the decaying weights) the values are described as:
    
    # mean_over_sliding_windows(zcorr_i(timeseries_1, timeseries_a) - z_corr_i(timeseries_2, timeseries_a))
    # std_over_sliding_windows(zcorr_i(timeseries_1, timeseries_a) - z_corr_i(timeseries_2, timeseries_a))
    # corr_over_sliding_windows(z_corr_i(timeseries_1, timeseries_a), z_corr_i(timeseries_2, timeseries_a))
    
    
    #The function will calculate nii_mean, nii_std, and nii_corr which is the 
    
    #Calculate window length and slide step length in number of TRs
    window_length = int(window_length_seconds/TR)
    slide_step_length = int(slide_step_seconds/TR)
    
    #Calculate the number of windows and make arrays
    #to store the correlations within each window
    num_steps = int(np.floor((len(timeseries_1) - window_length)/slide_step_length))
    corr_1 = np.zeros(num_steps)
    corr_2 = np.zeros(num_steps)
    
    #Calculate the tapered weights for the sliding windows
    weights = calc_tapered_weights(window_length, decay_constant)
    
    #Calculate the pearson product moment correlation
    #for each window
    for i in range(len(corr_1)):
        
        beginning = int(i*slide_step_length)
        end = int(i*slide_step_length+window_length)
        
        corr_1[i] = calc_pearson_product_moment_correlation(timeseries_1[beginning:end], timeseries_a[beginning:end], weights)
        corr_2[i] = calc_pearson_product_moment_correlation(timeseries_2[beginning:end], timeseries_a[beginning:end], weights)
    
    #Calculate fisher transformation
    z_corr_1 = np.arctanh(corr_1)
    z_corr_2 = np.arctanh(corr_2)
    
    #Calculate mean difference, std difference, and corr across all windows
    nii_mean = np.mean(np.subtract(z_corr_1, z_corr_2))
    nii_std = np.std(np.subtract(z_corr_1, z_corr_2))
    nii_corr = np.corrcoef(z_corr_1, z_corr_2)[1,0]
    
    return nii_mean, nii_std, nii_corr, np.mean(z_corr_1), np.mean(z_corr_2)
    
        

#This function calculates tapered weights for sliding-window
#analyses as described in "Time-Resolved Resting-State Brain
#Networks" by Zalesky et. al in PNAS 2014
def calc_tapered_weights(window_length, decay_constant = 0.333):
    
    #Calculates tapered weights for sliding window connectivity
    #analyses. Uses exponentially tapered window as defined in:
    #"Time-Resolved Resting-State Brain Networks" by Zalesky et. al
    #in PNAS 2014. Window length should be in number of TRs, and decay
    #constant should be relative to window_length. Default decay_constant
    #is 0.333, which is equivelant to one third the window_length.
    #Returns an array with the weights for each timepoint.
    
    #Decay constants << 1 will be more non-linear, and decay constants
    # >> 1 will approach linearity
    
    if decay_constant < 0:
        
        raise NameError('Error: Decay Constant must be positive')
    
    decay_constant = window_length*decay_constant
    w0 = (1 - np.exp(-1/decay_constant))/(1 - np.exp(-1*window_length/decay_constant))
    window_indices = np.linspace(1, window_length, window_length, dtype=int)
    weights = np.zeros(len(window_indices))
    for i in window_indices:
        weights[i - 1] = w0*np.exp((i - len(weights))/decay_constant)
        
    return weights




#These functions implement the pearson product-moment correlation
#as described in "Time-Resolved Resting-State Brain Networks" by 
#Zalesky et. al in PNAS 2014
def calc_weighted_mean(time_signal, weights):
    
    #print('Weighted_Mean: ' + str(np.sum(np.multiply(time_signal, weights))))
    return np.sum(np.multiply(time_signal, weights))/np.sum(weights)

def calc_weighted_cov(time_signal, weights):
    
    #Added a square term that is not listed in the supplement
    #because it makes more sense and I get an imaginary number otherwise
    
    x_dif = np.subtract(time_signal, calc_weighted_mean(time_signal, weights))
    weighted_x_dif = np.multiply(weights, x_dif)
    weighted_x_dif_sqr = np.power(weighted_x_dif, 2)
    weighted_cov = weighted_x_dif_sqr/np.sum(weights)
    #print('Weighted_std: ' + str(np.sqrt(np.sum(weighted_x_dif_sqr))))
    return weighted_cov

def calc_weighted_cov(time_signal_1, time_signal_2, weights):
    
    partial_1 = time_signal_1 - calc_weighted_mean(time_signal_1, weights)
    partial_2 = time_signal_2 - calc_weighted_mean(time_signal_2, weights)
    product = np.multiply(partial_1, partial_2)
    weighted_product = np.multiply(weights, product)
    #print('Weighted_cov: ' + str(np.sum(weighted_product)))
    return np.sum(weighted_product)

def calc_pearson_product_moment_correlation(time_signal_1, time_signal_2, weights):
    
    numerator = calc_weighted_cov(time_signal_1, time_signal_2, weights)
    denominator_sqrd = calc_weighted_cov(time_signal_1, time_signal_1, weights)*calc_weighted_cov(time_signal_2, time_signal_2, weights)
    denominator = np.sqrt(denominator_sqrd)
    return numerator/denominator



import numpy as np
import nibabel as nib




def convert_spherical_roi_coords_to_nifti(template_nifti_path, spherical_coords, radius, output_nifti_path, spherical_labels=None):
    
    #Template_nifti_path should point to a nifti with the desired
    #affine matrix/size, should be 3d, not a timeseries. Spherical_coords
    #should be a list of RAS coordinates for the spheres, radius should be a list
    #of radii for the different spheres, output_nifti_path is where the mask file
    #will be saved. Spherical labels is an optional list that can specify the number
    #assigned to values for different spheres. If this isn't set, spheres will be labeled
    #1, 2, 3 ... etc. 
    
    template_nifti = nib.load(template_nifti_path)
    affine = template_nifti.affine
    mask_vol = np.zeros(template_nifti.get_fdata().shape)
    
    for i in range(mask_vol.shape[0]):
        for j in range(mask_vol.shape[1]):
            for k in range(mask_vol.shape[2]):
                
                temp_ras = np.matmul(affine,[i, j, k, 1])[0:3]
                for l in range(len(spherical_coords)):

                    distance = np.linalg.norm(spherical_coords[l] - temp_ras)

                    if radius[l] <= distance:

                        if type(spherical_labels) == None:
                            mask_vol[i,j,k] = l + 1
                        else:
                            mask_vol[i,j,k] = spherical_labels[l]
                        break
                        
    template_header = template_nifti.header
    img = nib.nifti1Image(mask_vol, affine, header = template_header)
    nib.save(img, output_nifti_path)
    
    
    
    
    
    
def nifti_rois_to_time_signals(input_timeseries_nii_path, input_mask_nii_path, demedian_before_averaging = True):
    
    #Function that takes a 4d nifti file with path input_timeseries_nii_path,
    #and a 3d mask registered to the 4d timeseries (input_mask_nii_path) who has,
    #different values which specify different regions, and returns average time
    #signals for the different regions. By default, each voxel in a given ROI will be
    #divided by its median prior to averaging. The mean of all voxel medians will be
    #provided as output. To turn off this normalization, set demedian_before_averaging
    #to False.
    #
    #Output nifti_time_series - size n_regions, n_timepoints
    #unique_mask_vals - size n_regions (specifying the ID for each mask)
    #parc_mean_median_signal_intensities - size n_regions
    
    input_ts_nii = nib.load(input_timeseries_nii_path)
    input_mask_nii = nib.load(input_mask_nii_path)
    
    input_mask_matrix = input_mask_nii.get_fdata()
    input_ts_matrix = input_ts_nii.get_fdata()
    unique_mask_vals = np.unique(input_mask_matrix)
    unique_mask_vals.sort()
    unique_mask_vals = unique_mask_vals[1:]
    
    nifti_time_series = np.zeros((unique_mask_vals.shape[0], input_ts_matrix.shape[3]))
    parc_mean_median_signal_intensities = np.zeros(unique_mask_vals.shape[0])



    for i in range(len(unique_mask_vals)):

        inds = np.where(input_mask_matrix == unique_mask_vals[i])
        temp_timeseries = input_ts_matrix[inds]

        voxel_medians = np.nanmedian(temp_timeseries, axis=1)

        if demedian_before_averaging:
            temp_timeseries = temp_timeseries/voxel_medians[:,None]
            
        nifti_time_series[i,:] = np.nanmean(temp_timeseries, axis=0)
        parc_mean_median_signal_intensities[i] = np.nanmean(voxel_medians)
        
    return nifti_time_series, unique_mask_vals, parc_mean_median_signal_intensities



import os
import glob
import json
import numpy as np




def json_path_to_dict(path_to_json_file):
    
    with open(path_to_json_file,'r') as temp_file:
        
        json_contents = temp_file.read()
        dict_object = json.loads(json_contents)
        
    return dict_object


def save_dictionary(dictionary, path_for_dictionary_dir, overwrite = False):
    
        
    if path_for_dictionary_dir[-5:] == '.json':
        
        with open(os.path.join(path_for_dictionary_dir), 'w') as temp_file:
            json_dict = json.dumps(dictionary, indent = 4, sort_keys = True)
            temp_file.write(json_dict)        

    else:
        
        if os.path.exists(path_for_dictionary_dir) == True:
        
            if overwrite == False:

                raise NameError('Error: Parc Timeseries Dictionary Directory Already Exists at this path')

        else:
        
            os.makedirs(path_for_dictionary_dir)
        
        
        
        for temp_key in dictionary.keys():


            if type(dictionary[temp_key]) == dict:

                to_overwrite = overwrite
                save_dictionary(dictionary[temp_key], os.path.join(path_for_dictionary_dir, temp_key), overwrite = to_overwrite)

            else:


                if path_for_dictionary_dir[-3:] == 'txt':

                    with open(os.path.join(path_for_dictionary_dir, temp_key), 'w') as temp_file:
                        temp_file.write(str(dictionary[temp_key]))

                else:

                    np.save(os.path.join(path_for_dictionary_dir, temp_key), dictionary[temp_key])
                 
            
    return
                
                
    
    
    
def load_dictionary(dictionary_dir_path):
    
    dictionary = {}
    
    os.chdir(dictionary_dir_path)
    directory_contents = os.listdir()
    directory_final_name = dictionary_dir_path.split('/')[-1]
    for temp_file in directory_contents:
        
        if os.path.isdir(temp_file):
            
            dictionary[temp_file] = load_dictionary(os.path.join(dictionary_dir_path, temp_file))
            os.chdir(dictionary_dir_path)
            
        else:
            
            if temp_file[-5:] == '.json':
                                
                with open(temp_file, 'r') as temp_json:
                    json_contents = temp_json.read()
                    dictionary[temp_file] = json.loads(json_contents)
                    
            else:
            
                if dictionary_dir_path[-3:] == 'txt':

                    with open(temp_file,'r') as temp_reading:
                        file_contents = temp_reading.read()

                        try:
                            dictionary[temp_file.split('.')[0]] = float(file_contents)
                        except:

                            if file_contents[0] == '[' and file_contents[-1] == ']':
                                split_contents = file_contents[1:-1].split(',')
                                dictionary[temp_file.split('.')[0]] = []
                                for temp_item in split_contents:
                                    split_limited_1 = temp_item.replace(' ','')
                                    split_limited_2 = split_limited_1.replace("'","")
                                    dictionary[temp_file.split('.')[0]].append(split_limited_2)

                            else:
                                dictionary[temp_file.split('.')[0]] = file_contents                

                else:

                    dictionary[temp_file.split('.')[0]] = np.load(temp_file)
                
                
                
    return dictionary


def flatten_dictionary(dictionary):
    
    
    def inner_function(sub_dict, name_beginning):
        
        inner_dict = {}
        
        for temp_key in sub_dict.keys():
            
            if name_beginning != '':
                new_name_beginning = name_beginning + '_' + temp_key
            else:
                new_name_beginning = temp_key
                
            if type(sub_dict[temp_key]) == dict:
                
                new_dictionary = inner_function(sub_dict[temp_key], new_name_beginning)
                for temp_inner_key in new_dictionary.keys():
                    inner_dict[temp_inner_key] = new_dictionary[temp_inner_key]
                
            else:
                inner_dict[new_name_beginning] = sub_dict[temp_key]
                
        return inner_dict
    
    flattened_dictionary = inner_function(dictionary, '')
    return flattened_dictionary



import os
import glob
import json
import _pickle as pickle
from discovery_imaging_utils import imaging_utils
import pandas as pd
import numpy as np

import matplotlib.pyplot as plt
import scipy.interpolate as interp
from sklearn.decomposition import PCA
import scipy.interpolate as interp

def interpolate(timepoint_defined, signal, interp_type, TR):
    #defined_timepoints should be an array the length of the t with True at timepoints
    #that are defined and False at timepoints that are not defined. signal should also
    #be an array of length t. Timepoints at defined as False will be overwritten. This
    #script supports extrapolation at beginning/end of the time signal. As a quality control
    #for the spline interpolation, the most positive/negative values observed in the defined
    #portion of the signal are set as bounds for the interpolated signal
    
    #interpolation types supported:
    
        #(1) linear - takes closest point before/after undefined timepoint and interpolates.
        #    in end cases, uses the two points before/after
        #(2) cubic_spline - takes 5 closest time points before/after undefined timepoints
        #and applies cubic spline to undefined points. Uses defined signal to determine maximum/minimum
        #bounds for new interpolated points.
        #(3) spectral - yet to be implemented, will be based off of code from the 2014 Power
        #    paper

    timepoint_defined = np.array(timepoint_defined)
    
    true_inds = np.where(timepoint_defined == True)[0]
    false_inds = np.where(timepoint_defined == False)[0]
    

    signal_copy = np.array(signal)
    
    if interp_type == 'linear':
        
        #Still need to handle beginning/end cases
        
        for temp_timepoint in false_inds:
            
            
            #past_timepoint = true_inds[np.sort(np.where(true_inds < temp_timepoint)[0])[-1]]
            #future_timepoint = true_inds[np.sort(np.where(true_inds > temp_timepoint)[0])[0]]
            
            
            #Be sure there is at least one future timepoint and one past timepoint.
            #If there isn't, then grab either two past or two future timepoints and use those
            #for interpolation. If there aren't even two total past + future timepoints, then
            #just set the output to 0. Could also set the output to be unadjusted, but this
            #is a way to make the issue more obvious.
            temp_past_timepoint = np.sort(np.where(true_inds < temp_timepoint)[0])
            temp_future_timepoint = np.sort(np.where(true_inds > temp_timepoint)[0])
            
            #If we don't have enough data to interpolate/extrapolate
            if len(temp_past_timepoint) + len(temp_future_timepoint) < 2:
                
                signal_copy[temp_timepoint] = 0
               
            #If we do have enough data to interpolate/extrapolate
            else:
                
                if len(temp_past_timepoint) == 0:
                    past_timepoint = true_inds[temp_future_timepoint[1]]
                else:
                    past_timepoint = true_inds[temp_past_timepoint[-1]]

                if len(temp_future_timepoint) == 0:
                    future_timepoint = true_inds[temp_past_timepoint[-2]]
                else:
                    future_timepoint = true_inds[temp_future_timepoint[0]]

                #Find the appopriate past/future values
                past_value = signal_copy[int(past_timepoint)]
                future_value = signal_copy[int(future_timepoint)]

                #Use the interp1d function for interpolation
                interp_object = interp.interp1d([past_timepoint, future_timepoint], [past_value, future_value], bounds_error=False, fill_value='extrapolate')
                signal_copy[temp_timepoint] = interp_object(temp_timepoint).item(0)

        return signal_copy
            
    
    #For cubic spline interpolation, instead of taking the past/future timepoint
    #we will just take the closest 5 timepoints. If there aren't 5 timepoints, we will
    #set the output to 0
    if interp_type == 'cubic_spline':
        
        sorted_good = np.sort(signal_copy[true_inds])
        min_bound = sorted_good[0]
        max_bound = sorted_good[-1]
        
        #Continue if there are at least 5 good inds
        true_inds_needed = 5
        if len(true_inds) >= true_inds_needed:
        
            for temp_timepoint in false_inds:

                closest_inds = true_inds[np.argsort(np.absolute(true_inds - temp_timepoint))]
                closest_vals = signal_copy[closest_inds.astype(int)]
                interp_object = interp.interp1d(closest_inds, closest_vals, kind = 'cubic', bounds_error=False, fill_value='extrapolate')
                signal_copy[temp_timepoint.astype(int)] = interp_object(temp_timepoint).item(0)
                
            min_bound_exceded = np.where(signal_copy < min_bound)[0]
            if len(min_bound_exceded) > 0:
                
                signal_copy[min_bound_exceded] = min_bound
                
            max_bound_exceded = np.where(signal_copy > max_bound)[0]
            if len(max_bound_exceded) > 0:
                
                signal_copy[max_bound_exceded] = max_bound
             
        #If there aren't enough good timepoints, then set the bad timepoints = 0
        else:
        
            signal_copy[false_inds.astype(int)] = 0
                                               
                                               
        return signal_copy
    
    
    if interp_type == 'spectral':
        
        signal_copy = spectral_interpolation(timepoint_defined, signal_copy, TR)
        
        return signal_copy
    
                                               


def load_comps_dict(parc_obj, comps_dict):
    
    #Internal function to load a specific dictionary
    #file used in denoising. Supports multiple levels
    #of properties such as 'confounds.framewise_displacement'
    #and in cases where PCA reduction is used, does not include
    #n_skip_vols in PCA reduction, but pads the beginning of the
    #PCA reduction output with zeros to cover n_skip_vols.
    #
    # example_comps_dict = {'confounds.framewise_displacement' : False,
    #                       'confounds.twelve_motion_regs' : 3,
    #                       'aroma_noise_ics' : 3}
    #
    #This dictionary would form an output array <7,n_timepoints> including
    #framewise displacement, 3 PCs from twelve motion regressors, and 
    #3 PCs from the aroma noise ICs
    #
    #In cases where dim0 > 3.5*dim1 for an extracted element, swaps element dimensions
    
    if comps_dict == False:
        return False
    comps_matrix = []
        
    #Iterate through all key value pairs
    for key, value in comps_dict.items():
        
        #Load the current attribute of interest
        #if key has '.' representing multiple levels,
        #then recursively go through them to get the object
        if len(key.split('.')) == 1:
            
            temp_arr = getattr(parc_obj, key)
            
        else:
            
            levels = key.split('.')
            new_obj = getattr(parc_obj, levels[0])
            for temp_obj in levels[1:]:
                new_obj = getattr(new_obj, temp_obj)
            temp_arr = new_obj
        
        
        #If temp_arr is only 1d, at a second dimension for comparison
        if len(temp_arr.shape) == 1:
            
            temp_arr = np.reshape(temp_arr, (temp_arr.shape[0],1))
        
        #Current fix to reshape the aroma noise ICs... should
        #be addressing this at the parcel_timeseries object level though
        if temp_arr.shape[0] > 3.5*temp_arr.shape[1]:
            
            temp_arr = np.transpose(temp_arr)
        
        #If necessary, use PCA on the temp_arr
        if value != False:
            
            temp_arr = reduce_ics(temp_arr, value, parc_obj.n_skip_vols)
        
        #Either start a new array or stack to existing
        if comps_matrix == []:
        
            comps_matrix = temp_arr
                        
        else:
            
            comps_matrix = np.vstack((comps_matrix, temp_arr))
                    
    return comps_matrix
                                          

def reduce_ics(input_matrix, num_dimensions, n_skip_vols):

    #Takes input_matrix <num_original_dimensions, num_timepoints>. Returns
    #the num_dimensions top PCs from the input_matrix which are derived excluding
    #n_skip_vols, but zeros are padded to the beginning of the time series
    #in place of the n_skip_vols.
    
    
    if input_matrix.shape[0] > input_matrix.shape[1]:

        raise NameError('Error: input_matrix should have longer dim1 than dim0')
        
    if input_matrix.shape[0] <= 1:
        
        raise NameError('Error: input matrix must have multiple matrices')
        
    input_matrix_transposed = input_matrix.transpose()
    partial_input_matrix = input_matrix_transposed[n_skip_vols:,:]

    pca_temp = PCA()
    pca_temp.fit(partial_input_matrix)
    transformed_pcs = pca_temp.transform(partial_input_matrix)
    pca_time_signal = np.zeros((num_dimensions,input_matrix.shape[1]))
    pca_time_signal[:,n_skip_vols:] = transformed_pcs.transpose()[0:num_dimensions,:]

    #This section is from old iteration WITH ERROR!!!
    #good_components_inds = np.linspace(0,num_dimensions - 1, num = num_dimensions).astype(int)
    #pca_time_signal = np.zeros((num_dimensions, input_matrix.shape[1]))
    #pca_time_signal[:,n_skip_vols:] = pca_temp.components_[good_components_inds,:]
    
    return pca_time_signal


def demean_normalize(one_d_array):
    
    #Takes a 1d array and subtracts mean, and
    #divides by standard deviation
    
    temp_arr = one_d_array - np.nanmean(one_d_array)
        
    return temp_arr/np.nanstd(temp_arr)

def find_timepoints_to_scrub(parc_object, scrubbing_dictionary):
    
    #This function is an internal function for the main denoising script.
    #The purpose of this function is to return a array valued true for 
    #volumes to be included in subsequent analyses and a false for volumes
    #that need to be scrubbed.
    
    #This script will also get rid of the n_skip_vols at the beginning of the
    #scan. And these volumes don't get accounted for in Uniform.
    
    #If you don't want to scrub, just set scrubbing_dictionary equal to False, and
    #this script will only get rid of the initial volumes
    
    
    if type(scrubbing_dictionary) == type(False):
        
        if scrubbing_dictionary == False:
            
            temp_val = getattr(parc_object.confounds, 'framewise_displacement')
            good_arr = np.ones(temp_val.shape)
            good_arr[0:parc_object.n_skip_vols] = 0
            return good_arr
            
        else:
            
            raise NameError ('Error, if scrubbing dictionary is a boolean it must be False')
    
    
    if 'Uniform' in scrubbing_dictionary:
        
        amount_to_keep = scrubbing_dictionary.get('Uniform')[0]
        evaluation_metrics = scrubbing_dictionary.get('Uniform')[1]
        
        
        evaluation_array = []
        
        for temp_metric in evaluation_metrics:
                        
            if evaluation_array == []:
                
                evaluation_array = demean_normalize(getattr(parc_object.confounds, temp_metric))
                
            else:
                
                temp_val = np.absolute(demean_normalize(getattr(parc_object.confounds, temp_metric)))
                evaluation_array = np.add(evaluation_array, temp_val)
                
        num_timepoints_to_keep = int(evaluation_array.shape[0]*amount_to_keep)
        sorted_inds = np.argsort(evaluation_array)
        good_inds = sorted_inds[0:num_timepoints_to_keep]
        good_arr = np.zeros(evaluation_array.shape)
        good_arr[good_inds] = 1
        good_arr[0:parc_object.n_skip_vols] = 0
        
        return good_arr
    
    

    #If neither of the first two options were used, we will assume
    #they dictionary has appropriate key/value pairs describing scrubbing
    #criteria
    
    temp_val = getattr(parc_object.confounds, 'framewise_displacement')
    good_arr = np.ones(temp_val.shape)
    good_arr[0:parc_object.n_skip_vols] = 0
    
    #Iterate through all key/value pairs and set the good_arr
    #value for indices which the nuisance threshold is exceeded
    #equal to 0
    for temp_key, temp_thresh in scrubbing_dictionary.items():
        
        temp_values = getattr(parc_object.confounds, temp_key)
        bad_inds = np.where(temp_values > temp_thresh)[0]
        good_arr[bad_inds] = 0
        
    return good_arr
            
    
def spectral_interpolation(timepoint_defined, signal, TR):
    
    
    
    good_timepoint_inds = np.where(timepoint_defined == True)[0]
    bad_timepoint_inds = np.where(timepoint_defined == False)[0]
    num_timepoints = timepoint_defined.shape[0]
    signal_copy = signal.copy()
    
    t = float(TR)*good_timepoint_inds
    h = signal[good_timepoint_inds]
    TH = np.linspace(0,(num_timepoints - 1)*TR,num=num_timepoints)
    ofac = float(32)
    hifac = float(1)

    N = h.shape[0] #Number of timepoints
    T = np.max(t) - np.min(t) #Total observed timespan

    #Calculate sampling frequencies
    f = np.linspace(1/(T*ofac), hifac*N/(2*T), num = int(((hifac*N/(2*T))/((1/(T*ofac))) + 1)))

    #angular frequencies and constant offsets
    w = 2*np.pi*f


    t1 = np.reshape(t,((1,t.shape[0])))
    w1 = np.reshape(w,((w.shape[0],1)))

    tan_a = np.sum(np.sin(np.matmul(w1,t1*2)), axis=1)
    tan_b = np.sum(np.cos(np.matmul(w1,t1*2)), axis=1)
    tau = np.divide(np.arctan2(tan_a,tan_b),2*w)

    #Calculate the spectral power sine and cosine terms
    cterm = np.cos(np.matmul(w1,t1) - np.asarray([np.multiply(w,tau)]*t.shape[0]).transpose())
    sterm = np.sin(np.matmul(w1,t1) - np.asarray([np.multiply(w,tau)]*t.shape[0]).transpose())

    D = np.reshape(h,(1,h.shape[0]) )#This already has the correct shape

    ##C_final = (sum(Cmult,2).^2)./sum(Cterm.^2,2)
    #This calculation is done speerately for the numerator, denominator, and the division
    Cmult = np.multiply(cterm, D)
    numerator = np.sum(Cmult,axis=1)

    denominator = np.sum(np.power(cterm,2),axis=1)
    c = np.divide(numerator, denominator)

    #Repeat the above for sine term
    Smult = np.multiply(sterm,D)
    numerator = np.sum(Smult, axis=1)
    denominator = np.sum(np.power(sterm,2),axis=1)
    s = np.divide(numerator,denominator)

    #The inverse function to re-construct the original time series
    Time = TH
    T_rep = np.asarray([Time]*w.shape[0])
    #already have w defined
    prod = np.multiply(T_rep, w1)
    sin_t = np.sin(prod)
    cos_t = np.cos(prod)
    sw_p = np.multiply(sin_t,np.reshape(s,(s.shape[0],1)))
    cw_p = np.multiply(cos_t,np.reshape(c,(c.shape[0],1)))
    S = np.sum(sw_p,axis=0)
    C = np.sum(cw_p,axis=0)
    H = C + S

    #Normalize the reconstructed spectrum, needed when ofac > 1
    Std_H = np.std(H)
    Std_h = np.std(h)
    norm_fac = np.divide(Std_H,Std_h)
    H = np.divide(H,norm_fac)

    signal_copy[bad_timepoint_inds] = H[bad_timepoint_inds]
    
    return signal_copy
    
    

def flexible_denoise_parc(parc_obj, hpf_before_regression, scrub_criteria_dictionary, interpolation_method, noise_comps_dict, clean_comps_dict, high_pass, low_pass):
    

    #Function inputs:
    
    #parc_object = a parcellated timeseries object generated from 
    #file "imaging_utility_classes.py" which will contain both an
    #uncleaned parcellated time series, and other nuisance variables
    # etc. of interest
    
    #hpf_before_regression = the cutoff frequency for an optional high
    #pass filter that can be applied to the nuisance regressors (noise/clean) and the
    #uncleaned time signal before any regression or scrubbing occurs. Recommended
    #value would be 0.01 or False (False for if you want to skip this step)
    
    #scrub_criteria_dictionary = a dictionary that describes how scrubbing should be
    #implemented. Three main options are (1) instead of inputting a dictionary, setting this
    #variable to False, which will skip scrubbing, (2) {'Uniform' : [AMOUNT_TO_KEEP, ['std_dvars', 'framewise_displacement']]},
    #which will automatically only keep the best timepoints (for if you want all subjects to be scrubbed an equivelant amount).
    #This option will keep every timepoint if AMOUNT_TO_KEEP was 1, and no timepoints if it was 0. The list of confounds following
    #AMOUNT_TO_KEEP must at least contain one metric (but can be as many as you want) from parc_object.confounds. If more than one
    #metric is given, they will be z-transformed and their sum will be used to determine which timepoints should be
    #kept, with larger values being interpreted as noiser (WHICH MEANS THIS OPTION SHOULD ONLY BE USED WITH METRICS WHERE
    #ZERO OR NEGATIVE BASED VALUES ARE FINE AND LARGE POSITIVE VALUES ARE BAD) - this option could potentially produce
    #slightly different numbers of timepoints accross subjects still if the bad timepoints overlap to varying degrees with
    #the number of timepoints that are dropped at the beginning of the scan. (3) {'std_dvars' : 1.2, 'framewise_displacement' : 0.5} - 
    #similar to the "Uniform" option, the input metrics should be found in parc_object.confounds. Here only timepoints
    #with values below all specified thresholds will be kept for further analyses
    
    
    #interpolation_method: options are 'linear', 'cubic_spline' and (IN FUTURE) 'spectral'.
    #While scrubbed values are not included to determine any of the weights in the denoising
    #model, they will still be interpolated over and then "denoised" (have nuisance variance
    #removed) so that we have values to put into the optional filter at the end of processing.
    #The interpolated values only have any influence on the filtering proceedure, and will be again
    #removed from the time signal after filtering and thus not included in the final output. Interpolation
    #methods will do weird things if there aren't many timepoints after scrubbing. All interpolation
    #schemes besides spectral are essentially wrappers over scipy's 1d interpolation methods. 'spectral'
    #interpolation is implemented based on code from Anish Mitra/Jonathan Power
    #as shown in Power's 2014 NeuroImage paper
    
    
    #noise_comps_dict and clean_comps_dict both have the same syntax. The values
    #specified by both of these matrices will be used (along with constant and linear trend)
    #to construct the denoising regression model for the input timeseries, but only the
    #noise explained by the noise_comps_dict will be removed from the input timeseries (
    #plus also the constant and linear trend). Unlike the scrub_criteria_dictionary, the 
    #data specifed here do not need to come from the confounds section of the parc_object,
    #and because of this, if you want to include something found under parc_object.confounds,
    #you will need to specify "confounds" in the name. An example of the dictionary can be seen below:
    #
    #    clean_comps_dict = {'aroma_clean_ics' : False}
    #
    #
    #    noise_comps_dict = {'aroma_noise_ics' : 5,
    #                       'confounds.wmcsfgsr' : False
    #                       'confounds.twelve_motion_regs' : False
    #                        }
    #
    #
    #The dictionary key should specify an element to be included in the denoising process
    #and the dictionary value should be False if you don't want to do a PCA reduction on
    #the set of nuisance variables (this will be the case more often than not), alternatively
    #if the key represents a grouping of confounds, then you can use the value to specify the
    #number of principal components to kept from a reduction of the grouping. If hpf_before_regression
    #is used, the filtering will happen after the PCA.
    #
    #
    #
    
    
    #high_pass, low_pass: Filters to be applied as the last step in processing.
    #set as False if you don't want to use them, otherwise set equal to the 
    #cutoff frequency
    
    #
    #If any of the input parameters are set to True, they will be treated as if they were
    #set to False, because True values wouldn't mean anything....
    #
    #
    #
    #################################################################################################
    #################################################################################################
    #################################################################################################
    #################################################################################################
    #################################################################################################
    #################################################################################################
    
    #Create an array with 1s for timepoints to use, and 0s for scrubbed timepointsx
    good_timepoints = find_timepoints_to_scrub(parc_obj, scrub_criteria_dictionary)
    
    #Load the arrays with the data for both the clean and noise components to be used in regression
    clean_comps_pre_filter = load_comps_dict(parc_obj, clean_comps_dict)
    noise_comps_pre_filter = load_comps_dict(parc_obj, noise_comps_dict)
    
    #Apply an initial HPF to everything if necessary - this does not remove scrubbed timepoints,
    #but does skips the first n_skip_vols (which will be set to 0 and not used in subsequent steps)
    if hpf_before_regression != False:
        
        b, a = imaging_utils.construct_filter('highpass', [hpf_before_regression], parc_obj.TR, 6)
        
        #start with the clean comps matrix
        if type(clean_comps_pre_filter) != type(False):
                        
            clean_comps_post_filter = np.zeros(clean_comps_pre_filter.shape)
            for clean_dim in range(clean_comps_pre_filter.shape[0]):
                
                clean_comps_post_filter[clean_dim, parc_obj.n_skip_vols:] = imaging_utils.apply_filter(b, a, clean_comps_pre_filter[clean_dim, parc_obj.n_skip_vols:])
        
        #this option for both clean/noise indicates there is no input matrix to filter
        else:
            
            clean_comps_post_filter = False
         
        #Move to the noise comps matrix
        if type(noise_comps_pre_filter) != type(False):
            
            noise_comps_post_filter = np.zeros(noise_comps_pre_filter.shape)
            for noise_dim in range(noise_comps_pre_filter.shape[0]):
                                
                noise_comps_post_filter[noise_dim, parc_obj.n_skip_vols:] = imaging_utils.apply_filter(b, a, noise_comps_pre_filter[noise_dim, parc_obj.n_skip_vols:])
        
        else:
            
            noise_comps_post_filter = False
         
        #then filter the original time signal
        filtered_time_series = np.zeros(parc_obj.time_series.shape)
        for original_ts_dim in range(parc_obj.time_series.shape[0]):
            
            filtered_time_series[original_ts_dim, parc_obj.n_skip_vols:] = imaging_utils.apply_filter(b, a, parc_obj.time_series[original_ts_dim, parc_obj.n_skip_vols:])
    
    #If you don't want to apply the initial HPF, then
    #just make a copy of the matrices of interest
    else:
        
        clean_comps_post_filter = clean_comps_pre_filter
        noise_comps_post_filter = noise_comps_pre_filter
        filtered_time_series = parc_obj.time_series
        
    
    
    
    #Now create the nuisance regression model. Only do this step if
    #the noise_comps_post_filter isn't false.
    good_timepoint_inds = np.where(good_timepoints == True)[0]
    bad_timepoint_inds = np.where(good_timepoints == False)[0]
    if type(noise_comps_post_filter) == type(False):
        
        regressed_time_signal = filtered_time_series
        
    else:
        
        
        #Weird thing where I need to swap dimensions here...(implemented correctly)      
        
        #First add constant/linear trend to the denoising model
        constant = np.ones((1,filtered_time_series.shape[1]))
        linear_trend = np.linspace(0,filtered_time_series.shape[1],num=filtered_time_series.shape[1])
        linear_trend = np.reshape(linear_trend, (1,filtered_time_series.shape[1]))[0]
        noise_comps_post_filter = np.vstack((constant, linear_trend, noise_comps_post_filter))
        
        regressed_time_signal = np.zeros(filtered_time_series.shape).transpose()
        filtered_time_series_T = filtered_time_series.transpose()
        
        #If there aren't any clean components,
        #do a "hard" or "agressive" denosing
        if type(clean_comps_post_filter) == type(False):
                   
            noise_comps_post_filter_T_to_be_used = noise_comps_post_filter[:,good_timepoint_inds].transpose()
            XT_X_Neg1_XT = imaging_utils.calculate_XT_X_Neg1_XT(noise_comps_post_filter_T_to_be_used)
            for temp_time_signal_dim in range(filtered_time_series.shape[0]):
                regressed_time_signal[good_timepoint_inds,temp_time_signal_dim] = imaging_utils.partial_clean_fast(filtered_time_series_T[good_timepoint_inds,temp_time_signal_dim], XT_X_Neg1_XT, noise_comps_post_filter_T_to_be_used)

                        
        
        #If there are clean components, then
        #do a "soft" denoising
        else:
    
            full_matrix_to_be_used = np.vstack((noise_comps_post_filter, clean_comps_post_filter))[:,good_timepoint_inds].transpose()
            noise_comps_post_filter_T_to_be_used = noise_comps_post_filter[:,good_timepoint_inds].transpose()
            XT_X_Neg1_XT = imaging_utils.calculate_XT_X_Neg1_XT(full_matrix_to_be_used)
                        
            for temp_time_signal_dim in range(filtered_time_series.shape[0]):
                regressed_time_signal[good_timepoint_inds,temp_time_signal_dim] = imaging_utils.partial_clean_fast(filtered_time_series_T[good_timepoint_inds,temp_time_signal_dim], XT_X_Neg1_XT, noise_comps_post_filter_T_to_be_used)
                                
        
        #Put back into original dimensions
        regressed_time_signal = regressed_time_signal.transpose()
        
        
    #Now apply interpolation
    interpolated_time_signal = np.zeros(regressed_time_signal.shape)
    
    if interpolation_method == 'spectral':
        
        interpolated_time_signal = spectral_interpolation_fast(good_timepoints, regressed_time_signal, parc_obj.TR)
        
    else:
        for dim in range(regressed_time_signal.shape[0]):
            interpolated_time_signal[dim,:] = interpolate(good_timepoints, regressed_time_signal[dim,:], interpolation_method, parc_obj.TR)

    #Now if necessary, apply additional filterign:
    if high_pass == False and low_pass == False:

        filtered_time_signal = interpolated_time_signal

    else:

        if high_pass != False and low_pass == False:

            b, a = imaging_utils.construct_filter('highpass', [high_pass], parc_obj.TR, 6)

        elif high_pass == False and low_pass != False:

            b, a = imaging_utils.construct_filter('lowpass', [low_pass], parc_obj.TR, 6)

        elif high_pass != False and low_pass != False:

            b, a = imaging_utils.construct_filter('bandpass', [high_pass, low_pass], parc_obj.TR, 6)

        filtered_time_signal = np.zeros(regressed_time_signal.shape)
        for dim in range(regressed_time_signal.shape[0]):

            filtered_time_signal[dim,:] = imaging_utils.apply_filter(b,a,regressed_time_signal[dim,:])
                
        
    #Now set all the undefined timepoints to Nan
    cleaned_time_signal = filtered_time_signal
    cleaned_time_signal[:,bad_timepoint_inds] = np.nan
    
    return cleaned_time_signal, good_timepoint_inds
    
    
    
def flexible_orth_denoise_parc(parc_obj, hpf_before_regression, scrub_criteria_dictionary, interpolation_method, noise_comps_dict, clean_comps_dict, high_pass, low_pass):
    
    #THIS FUNCTION IS THE SAME AS FLEXIBLE DENOISE PARC,
    #EXCEPT FOR HERE, THE REGRESSORS IDENTIFIED BY CLEAN
    #COMPS DICT ARE REGRESSED FROM THE REGRESSORS IDENTIFIED
    #BY NOISE COMPS DICT PRIOR TO THE REGRESSORS FROM NOISE COMPS
    #DICT BEING USED TO CLEAN THE TIMESERIES. THIS MEANS THE MODEL
    #TO CLEAN THE TIMESERIES WILL ONLY CONTAIN THE ORTHOGONALIZED
    #NUISANCE VARIABLES (filtering and other options will be applied
    #as per usual)

    #Function inputs:
    
    #parc_object = a parcellated timeseries object generated from 
    #file "imaging_utility_classes.py" which will contain both an
    #uncleaned parcellated time series, and other nuisance variables
    # etc. of interest
    
    #hpf_before_regression = the cutoff frequency for an optional high
    #pass filter that can be applied to the nuisance regressors (noise/clean) and the
    #uncleaned time signal before any regression or scrubbing occurs. Recommended
    #value would be 0.01 or False (False for if you want to skip this step)
    
    #scrub_criteria_dictionary = a dictionary that describes how scrubbing should be
    #implemented. Three main options are (1) instead of inputting a dictionary, setting this
    #variable to False, which will skip scrubbing, (2) {'Uniform' : [AMOUNT_TO_KEEP, ['std_dvars', 'framewise_displacement']]},
    #which will automatically only keep the best timepoints (for if you want all subjects to be scrubbed an equivelant amount).
    #This option will keep every timepoint if AMOUNT_TO_KEEP was 1, and no timepoints if it was 0. The list of confounds following
    #AMOUNT_TO_KEEP must at least contain one metric (but can be as many as you want) from parc_object.confounds. If more than one
    #metric is given, they will be z-transformed and their sum will be used to determine which timepoints should be
    #kept, with larger values being interpreted as noiser (WHICH MEANS THIS OPTION SHOULD ONLY BE USED WITH METRICS WHERE
    #ZERO OR NEGATIVE BASED VALUES ARE FINE AND LARGE POSITIVE VALUES ARE BAD) - this option could potentially produce
    #slightly different numbers of timepoints accross subjects still if the bad timepoints overlap to varying degrees with
    #the number of timepoints that are dropped at the beginning of the scan. (3) {'std_dvars' : 1.2, 'framewise_displacement' : 0.5} - 
    #similar to the "Uniform" option, the input metrics should be found in parc_object.confounds. Here only timepoints
    #with values below all specified thresholds will be kept for further analyses
    
    
    #interpolation_method: options are 'linear', 'cubic_spline' and (IN FUTURE) 'spectral'.
    #While scrubbed values are not included to determine any of the weights in the denoising
    #model, they will still be interpolated over and then "denoised" (have nuisance variance
    #removed) so that we have values to put into the optional filter at the end of processing.
    #The interpolated values only have any influence on the filtering proceedure, and will be again
    #removed from the time signal after filtering and thus not included in the final output. Interpolation
    #methods will do weird things if there aren't many timepoints after scrubbing. All interpolation
    #schemes besides spectral are essentially wrappers over scipy's 1d interpolation methods. 'spectral'
    #interpolation is implemented based on code from Anish Mitra/Jonathan Power
    #as shown in Power's 2014 NeuroImage paper
    
    
    #noise_comps_dict and clean_comps_dict both have the same syntax. The values
    #specified by both of these matrices will be used (along with constant and linear trend)
    #to construct the denoising regression model for the input timeseries, but only the
    #noise explained by the noise_comps_dict will be removed from the input timeseries (
    #plus also the constant and linear trend). Unlike the scrub_criteria_dictionary, the 
    #data specifed here do not need to come from the confounds section of the parc_object,
    #and because of this, if you want to include something found under parc_object.confounds,
    #you will need to specify "confounds" in the name. An example of the dictionary can be seen below:
    #
    #    clean_comps_dict = {'aroma_clean_ics' : False}
    #
    #
    #    noise_comps_dict = {'aroma_noise_ics' : 5,
    #                       'confounds.wmcsfgsr' : False
    #                       'confounds.twelve_motion_regs' : False
    #                        }
    #
    #
    #The dictionary key should specify an element to be included in the denoising process
    #and the dictionary value should be False if you don't want to do a PCA reduction on
    #the set of nuisance variables (this will be the case more often than not), alternatively
    #if the key represents a grouping of confounds, then you can use the value to specify the
    #number of principal components to kept from a reduction of the grouping. If hpf_before_regression
    #is used, the filtering will happen after the PCA.
    #
    #
    #
    
    
    #high_pass, low_pass: Filters to be applied as the last step in processing.
    #set as False if you don't want to use them, otherwise set equal to the 
    #cutoff frequency
    
    #
    #If any of the input parameters are set to True, they will be treated as if they were
    #set to False, because True values wouldn't mean anything....
    #
    #
    #
    #################################################################################################
    #################################################################################################
    #################################################################################################
    #################################################################################################
    #################################################################################################
    #################################################################################################
    
    #Create an array with 1s for timepoints to use, and 0s for scrubbed timepointsx
    good_timepoints = find_timepoints_to_scrub(parc_obj, scrub_criteria_dictionary)
    
    #Load the arrays with the data for both the clean and noise components to be used in regression
    clean_comps_pre_filter = load_comps_dict(parc_obj, clean_comps_dict)
    noise_comps_pre_filter = load_comps_dict(parc_obj, noise_comps_dict)
    
    #Apply an initial HPF to everything if necessary - this does not remove scrubbed timepoints,
    #but does skips the first n_skip_vols (which will be set to 0 and not used in subsequent steps)
    if hpf_before_regression != False:
        
        b, a = imaging_utils.construct_filter('highpass', [hpf_before_regression], parc_obj.TR, 6)
        
        #start with the clean comps matrix
        if type(clean_comps_pre_filter) != type(False):
                        
            clean_comps_post_filter = np.zeros(clean_comps_pre_filter.shape)
            for clean_dim in range(clean_comps_pre_filter.shape[0]):
                
                clean_comps_post_filter[clean_dim, parc_obj.n_skip_vols:] = imaging_utils.apply_filter(b, a, clean_comps_pre_filter[clean_dim, parc_obj.n_skip_vols:])
        
        #this option for both clean/noise indicates there is no input matrix to filter
        else:
            
            clean_comps_post_filter = False
         
        #Move to the noise comps matrix
        if type(noise_comps_pre_filter) != type(False):
            
            noise_comps_post_filter = np.zeros(noise_comps_pre_filter.shape)
            for noise_dim in range(noise_comps_pre_filter.shape[0]):
                                
                noise_comps_post_filter[noise_dim, parc_obj.n_skip_vols:] = imaging_utils.apply_filter(b, a, noise_comps_pre_filter[noise_dim, parc_obj.n_skip_vols:])
        
        else:
            
            noise_comps_post_filter = False
         
        #then filter the original time signal
        filtered_time_series = np.zeros(parc_obj.time_series.shape)
        for original_ts_dim in range(parc_obj.time_series.shape[0]):
            
            filtered_time_series[original_ts_dim, parc_obj.n_skip_vols:] = imaging_utils.apply_filter(b, a, parc_obj.time_series[original_ts_dim, parc_obj.n_skip_vols:])
    
    #If you don't want to apply the initial HPF, then
    #just make a copy of the matrices of interest
    else:
        
        clean_comps_post_filter = clean_comps_pre_filter
        noise_comps_post_filter = noise_comps_pre_filter
        filtered_time_series = parc_obj.time_series
        
    
    
    
    #Now create the nuisance regression model. Only do this step if
    #the noise_comps_post_filter isn't false.
    good_timepoint_inds = np.where(good_timepoints == True)[0]
    bad_timepoint_inds = np.where(good_timepoints == False)[0]
    if type(noise_comps_post_filter) == type(False):
        
        regressed_time_signal = filtered_time_series
        
    else:
        
        
        #Weird thing where I need to swap dimensions here...(implemented correctly)      
        
        #First add constant/linear trend to the denoising model
        constant = np.ones((1,filtered_time_series.shape[1]))
        linear_trend = np.linspace(0,filtered_time_series.shape[1],num=filtered_time_series.shape[1])
        linear_trend = np.reshape(linear_trend, (1,filtered_time_series.shape[1]))[0]
        noise_comps_post_filter = np.vstack((constant, linear_trend, noise_comps_post_filter))
        
        regressed_time_signal = np.zeros(filtered_time_series.shape).transpose()
        filtered_time_series_T = filtered_time_series.transpose()
        
        #If there aren't any clean components,
        #do a "hard" or "agressive" denosing
        if type(clean_comps_post_filter) == type(False):
                   
            noise_comps_post_filter_T_to_be_used = noise_comps_post_filter[:,good_timepoint_inds].transpose()
            XT_X_Neg1_XT = imaging_utils.calculate_XT_X_Neg1_XT(noise_comps_post_filter_T_to_be_used)
            for temp_time_signal_dim in range(filtered_time_series.shape[0]):
                regressed_time_signal[good_timepoint_inds,temp_time_signal_dim] = imaging_utils.partial_clean_fast(filtered_time_series_T[good_timepoint_inds,temp_time_signal_dim], XT_X_Neg1_XT, noise_comps_post_filter_T_to_be_used)

                        
        
        #If there are clean components, then
        #do a "soft" denoising...
        
        
###########################################################################
###########################################################################
####THIS CHUNK OF CODE IS THE ONLY THING TO BE CHANGED BETWEEN#############
####THE ORIGINAL FLEXIBLE DENOISE FUNC FUNCTION AND THIS ONE###############
###########################################################################
###########################################################################
        
        else:
    
            noise_comps_post_filter_T_to_be_used = noise_comps_post_filter[:,good_timepoint_inds].transpose()
            clean_comps_post_filter_T_to_be_used = clean_comps_post_filter[:,good_timepoint_inds].transpose()

            orth_noise_comps_post_filter = np.zeros(noise_comps_post_filter.shape).transpose()

            initial_XT_X_Neg1_XT = imaging_utils.calculate_XT_X_Neg1_XT(clean_comps_post_filter_T_to_be_used)
            for temp_time_signal_dim in range(orth_noise_comps_post_filter.shape[1]):
                orth_noise_comps_post_filter[good_timepoint_inds,temp_time_signal_dim] = imaging_utils.partial_clean_fast(noise_comps_post_filter_T_to_be_used[:,temp_time_signal_dim], initial_XT_X_Neg1_XT, clean_comps_post_filter_T_to_be_used)

            noise_comps_post_filter_T_to_be_used = orth_noise_comps_post_filter[good_timepoint_inds,:]
            XT_X_Neg1_XT = imaging_utils.calculate_XT_X_Neg1_XT(noise_comps_post_filter_T_to_be_used)
            for temp_time_signal_dim in range(filtered_time_series.shape[0]):
                regressed_time_signal[good_timepoint_inds,temp_time_signal_dim] = imaging_utils.partial_clean_fast(filtered_time_series_T[good_timepoint_inds,temp_time_signal_dim], XT_X_Neg1_XT, noise_comps_post_filter_T_to_be_used)
                
            #full_matrix_to_be_used = np.vstack((noise_comps_post_filter, clean_comps_post_filter))[:,good_timepoint_inds].transpose()
            #noise_comps_post_filter_T_to_be_used = noise_comps_post_filter[:,good_timepoint_inds].transpose()
            #XT_X_Neg1_XT = imaging_utils.calculate_XT_X_Neg1_XT(full_matrix_to_be_used)
                        
            #for temp_time_signal_dim in range(filtered_time_series.shape[0]):
            #    regressed_time_signal[good_timepoint_inds,temp_time_signal_dim] = imaging_utils.partial_clean_fast(filtered_time_series_T[good_timepoint_inds,temp_time_signal_dim], XT_X_Neg1_XT, noise_comps_post_filter_T_to_be_used)
                                
        
        #Put back into original dimensions
        regressed_time_signal = regressed_time_signal.transpose()
        
        
###########################################################################
###########################################################################
###########################################################################
###########################################################################
###########################################################################
        
        
    #Now apply interpolation
    interpolated_time_signal = np.zeros(regressed_time_signal.shape)
    
    if interpolation_method == 'spectral':
        
        interpolated_time_signal = spectral_interpolation_fast(good_timepoints, regressed_time_signal, parc_obj.TR)
        
    else:
        for dim in range(regressed_time_signal.shape[0]):
            interpolated_time_signal[dim,:] = interpolate(good_timepoints, regressed_time_signal[dim,:], interpolation_method, parc_obj.TR)

    #Now if necessary, apply additional filterign:
    if high_pass == False and low_pass == False:

        filtered_time_signal = interpolated_time_signal

    else:

        if high_pass != False and low_pass == False:

            b, a = imaging_utils.construct_filter('highpass', [high_pass], parc_obj.TR, 6)

        elif high_pass == False and low_pass != False:

            b, a = imaging_utils.construct_filter('lowpass', [low_pass], parc_obj.TR, 6)

        elif high_pass != False and low_pass != False:

            b, a = imaging_utils.construct_filter('bandpass', [high_pass, low_pass], parc_obj.TR, 6)

        filtered_time_signal = np.zeros(regressed_time_signal.shape)
        for dim in range(regressed_time_signal.shape[0]):

            filtered_time_signal[dim,:] = imaging_utils.apply_filter(b,a,regressed_time_signal[dim,:])
                
        
    #Now set all the undefined timepoints to Nan
    cleaned_time_signal = filtered_time_signal
    cleaned_time_signal[:,bad_timepoint_inds] = np.nan
    
    return cleaned_time_signal, good_timepoint_inds
    
    
    
def spectral_interpolation_fast(timepoint_defined, signal, TR):
    
    
    good_timepoint_inds = np.where(timepoint_defined == True)[0]
    bad_timepoint_inds = np.where(timepoint_defined == False)[0]
    num_timepoints = timepoint_defined.shape[0]
    signal_copy = signal.copy()
    
    t = float(TR)*good_timepoint_inds
    h = signal[:,good_timepoint_inds]
    TH = np.linspace(0,(num_timepoints - 1)*TR,num=num_timepoints)
    ofac = float(8) #Higher than this is slow without good quality improvements
    hifac = float(1)

    N = timepoint_defined.shape[0] #Number of timepoints
    T = np.max(t) - np.min(t) #Total observed timespan

    #Calculate sampling frequencies
    f = np.linspace(1/(T*ofac), hifac*N/(2*T), num = int(((hifac*N/(2*T))/((1/(T*ofac))) + 1)))

    #angular frequencies and constant offsets
    w = 2*np.pi*f

    t1 = np.reshape(t,((1,t.shape[0])))
    w1 = np.reshape(w,((w.shape[0],1)))

    tan_a = np.sum(np.sin(np.matmul(w1,t1*2)), axis=1)
    tan_b = np.sum(np.cos(np.matmul(w1,t1*2)), axis=1)
    tau = np.divide(np.arctan2(tan_a,tan_b),2*w)

    a1 = np.matmul(w1,t1)
    b1 = np.asarray([np.multiply(w,tau)]*t.shape[0]).transpose()
    cs_input = a1 - b1

    #Calculate the spectral power sine and cosine terms
    cterm = np.cos(cs_input)
    sterm = np.sin(cs_input)
    
    cos_denominator = np.sum(np.power(cterm,2),axis=1)
    sin_denominator = np.sum(np.power(sterm,2),axis=1)
    
    #The inverse function to re-construct the original time series pt. 1
    Time = TH
    T_rep = np.asarray([Time]*w.shape[0])
    #already have w defined
    prod = np.multiply(T_rep, w1)
    sin_t = np.sin(prod)
    cos_t = np.cos(prod)
    
    for i in range(h.shape[0]):

        ##C_final = (sum(Cmult,2).^2)./sum(Cterm.^2,2)
        #This calculation is done speerately for the numerator, denominator, and the division
        Cmult = np.multiply(cterm, h[i,:])
        numerator = np.sum(Cmult,axis=1)

        c = np.divide(numerator, cos_denominator)

        #Repeat the above for sine term
        Smult = np.multiply(sterm,h[i,:])
        numerator = np.sum(Smult, axis=1)
        s = np.divide(numerator,sin_denominator)

        #The inverse function to re-construct the original time series pt. 2
        sw_p = np.multiply(sin_t,np.reshape(s,(s.shape[0],1)))
        cw_p = np.multiply(cos_t,np.reshape(c,(c.shape[0],1)))

        S = np.sum(sw_p,axis=0)
        C = np.sum(cw_p,axis=0)
        H = C + S

        #Normalize the reconstructed spectrum, needed when ofac > 1
        Std_H = np.std(H)
        Std_h = np.std(h)
        norm_fac = np.divide(Std_H,Std_h)
        H = np.divide(H,norm_fac)

        signal_copy[i,bad_timepoint_inds] = H[bad_timepoint_inds]

    
    return signal_copy





import pandas as pd
import numpy as np
import os
import glob
from discovery_imaging_utils import imaging_utils
from discovery_imaging_utils import nifti_utils
import json

#Run in this order:
#(1) file_paths_dict = generate_file_paths(....)
#(2) check if files present with all_file_paths_exist(...)
#(3) if files present, parc_ts_dict = populate_parc_dictionary(....)
#(4) then use save/load functions to store in directory structure




def all_file_paths_exist(file_path_dictionary):
    
    #Check if all files exist, and if they don't
    #return False
    files_present = True
    
    for temp_field in file_path_dictionary:
        
        if os.path.exists(file_path_dictionary[temp_field]) == False:
            print(file_path_dictionary[temp_field])
            files_present = False
            
    return files_present

def generate_file_paths(lh_gii_path=None, lh_parcellation_path=None, nifti_ts_path=None,
                         nifti_parcellation_path=None, aroma_included=True):
    
    #This function gathers paths useful for imaging analyses... either a gifti or nifti
    #or both must be specified along with accompanying parcellation paths, outputs a dictionary
    #with relevant paths, output of this function can be used with "all_file_paths_exist", then
    #with "populate_parc_dictionary", then can be saved with "save_dictionary"
    
    path_dictionary = {}
    prefix = ''
    
    #Gifti related paths
    if type(lh_gii_path) != type(None):
        
        if type(lh_parcellation_path) == type(None):
            
            raise NameError('Error: need to specify lh parcellation path if you specify a lh_gii_path')
        
        path_dictionary['lh_func_path'] = lh_gii_path
        path_dictionary['rh_func_path'] = lh_gii_path[:-10] + 'R.func.gii'
        path_dictionary['lh_parcellation_path'] = lh_parcellation_path
        
        lh_parcel_end = lh_parcellation_path.split('/')[-1]
    
        path_dictionary['rh_parcellation_path'] = lh_parcellation_path[:-len(lh_parcel_end)] + 'r' + lh_parcellation_path.split('/')[-1][1:]

        #For finding aroma/nuisance files
        prefix = '_'.join(lh_gii_path.split('_')[:-2])
    
    #Nifti related paths
    if type(nifti_ts_path) != type(None):
        
        if type(nifti_parcellation_path) == type(None):
            
            raise NameError('Error: need to specify nifti parcellation path if you specify a nifti_ts_path')
            
        path_dictionary['nifti_func_path'] = nifti_ts_path
        path_dictionary['nifti_parcellation_path'] = nifti_parcellation_path
        
        nifti_parcellation_json = nifti_parcellation_path.split('.')[0] + '.json'
        if os.path.exists(nifti_parcellation_json) == True:
            path_dictionary['nifti_parcellation_json_path'] = nifti_parcellation_json
        
        if prefix != '':
            if '_'.join(nifti_ts_path.split('_')[:-3]) != prefix:
                raise NameError('Error: It doesnt look like the nifti and gifti timeseries point to the same run')
        
        prefix = '_'.join(nifti_ts_path.split('_')[:-3])
     
    
    #Aroma related paths
    if aroma_included:
        path_dictionary['melodic_mixing_path'] = prefix + '_desc-MELODIC_mixing.tsv'
        path_dictionary['aroma_noise_ics_path'] = prefix + '_AROMAnoiseICs.csv'

    #Confounds path  
    path_dictionary['confounds_regressors_path'] = prefix + '_desc-confounds_regressors.tsv'
    
    return path_dictionary
        

def populate_parc_dictionary(file_path_dictionary, TR):
    
    parc_ts_dictionary = {}
    has_gifti = False
    has_nifti = False
    
    ##############################################################################
    #Load the timeseries data and apply parcellation, saving also the parcel labels
    
    if 'lh_func_path' in file_path_dictionary.keys():
    
        has_gifti = True
        lh_data = imaging_utils.load_gifti_func(file_path_dictionary['lh_func_path'])
        rh_data = imaging_utils.load_gifti_func(file_path_dictionary['rh_func_path'])
        time_series_cortex, parcel_labels_cortex, parc_median_signal_intensities = imaging_utils.demedian_parcellate_func_combine_hemis(lh_data, rh_data, file_path_dictionary['lh_parcellation_path'], file_path_dictionary['rh_parcellation_path'])

        parc_ts_dictionary['labels_surface'] = parcel_labels_cortex
        parc_ts_dictionary['time_series_surface'] = time_series_cortex
        parc_ts_dictionary['median_ts_intensities_surface'] = parc_median_signal_intensities
        
        
        
    if 'nifti_func_path' in file_path_dictionary.keys():
            
        
        has_nifti = True
        time_series_nifti, parcel_labels_nifti, parc_median_signal_intensities_nifti = nifti_utils.nifti_rois_to_time_signals(file_path_dictionary['nifti_func_path'], file_path_dictionary['nifti_parcellation_path'], demedian_before_averaging = True)

        
        if 'nifti_parcellation_json_path' in file_path_dictionary.keys():
            
            with open(file_path_dictionary['nifti_parcellation_json_path'], 'r') as temp_file:
                
                json_contents = temp_file.read()
                json_dict = json.loads(json_contents)
            
            
            parc_ts_dictionary['nifti_parcellation_info.json'] = json_dict
            
            output_labels = []
            for temp_label in parcel_labels_nifti:
            
                strint_label = str(int(temp_label))
                
                if 'unique_region_identifier' in json_dict[strint_label].keys():
                    
                    output_labels.append(json_dict[strint_label]['unique_region_identifier'])
                    
                else:
                    
                    output_labels.append(strint_label)
                    
            parcel_labels_nifti = output_labels
            
        parc_ts_dictionary['labels_nifti'] = parcel_labels_nifti
        parc_ts_dictionary['time_series_nifti'] = time_series_nifti
        parc_ts_dictionary['median_ts_intensities_nifti'] = parc_median_signal_intensities_nifti
        
        
    if (has_nifti and has_gifti):
        
        parc_ts_dictionary['time_series'] = np.vstack((parc_ts_dictionary['time_series_surface'],parc_ts_dictionary['time_series_nifti']))
        parc_ts_dictionary['labels'] = parc_ts_dictionary['labels_surface'] + parc_ts_dictionary['labels_nifti']
        parc_ts_dictionary['median_ts_intensities'] = np.hstack((parc_ts_dictionary['median_ts_intensities_surface'],parc_ts_dictionary['median_ts_intensities_nifti']))
    
    elif has_nifti:
        
        parc_ts_dictionary['time_series'] = parc_ts_dictionary['time_series_nifti']
        parc_ts_dictionary['labels'] = parc_ts_dictionary['labels_nifti']
        parc_ts_dictionary['median_ts_intensities'] = parc_ts_dictionary['median_ts_intensities_nifti']
        
    elif has_gifti:
        
        parc_ts_dictionary['time_series'] = parc_ts_dictionary['time_series_surface']
        parc_ts_dictionary['labels'] = parc_ts_dictionary['labels_surface']
        parc_ts_dictionary['median_ts_intensities'] = parc_ts_dictionary['median_ts_intensities_surface']
        
    else:
        
        raise NameError('Error: File Path Dictionary must either have gifti or nifti specified')
    
    
    parc_ts_dictionary['TR'] = TR
    
    
    
    ####################################################
    #Load the melodic IC time series
    melodic_df = pd.read_csv(file_path_dictionary['melodic_mixing_path'], sep="\t", header=None)
    parc_ts_dictionary['melodic_mixing_matrix'] = melodic_df.values 

    ####################################################
    #Load the indices of the aroma ics
    aroma_ics_df = pd.read_csv(file_path_dictionary['aroma_noise_ics_path'], header=None)
    parc_ts_dictionary['aroma_noise_ic_inds'] = (aroma_ics_df.values - 1).reshape(-1,1)
    

    ####################################################
    #Gather the ICs identified as noise/clean by AROMA
    noise_comps = parc_ts_dictionary['aroma_noise_ic_inds']
    all_ics = melodic_df.values 
    mask = np.zeros(all_ics.shape[1],dtype=bool)
    mask[noise_comps] = True
    parc_ts_dictionary['aroma_noise_ic_timeseries'] = all_ics[:,~mask]
    parc_ts_dictionary['aroma_clean_ic_timeseries'] = all_ics[:,mask]

    ####################################################
    #Get the variables from the confounds regressors file
    #confound_df = pd.read_csv(self.confounds_regressors_path, sep='\t')
    #for (columnName, columnData) in confound_df.iteritems():
    #    setattr(self, columnName, columnData.as_matrix())
    parc_ts_dictionary['confounds'] = populate_confounds_dict(file_path_dictionary)
    
    parc_ts_dictionary['file_path_dictionary.json'] = file_path_dictionary
    
    ###################################################
    #Calculate general info, such as number of high motion timepoints
    #number of volumes that need to be skipped at the beginning of the
    #scan, etc.
    
    parc_ts_dictionary['general_info.json'] = populate_general_info_dict(parc_ts_dictionary)
    
    
    
    return parc_ts_dictionary
    

def populate_general_info_dict(parc_ts_dictionary):
    
    general_info_dict = {}
    
    ###################################################
    #Calculate the number of timepoints to skip at the beginning for this person.
    #If equal to zero, we will actually call it one so that we don't run into any
    #issues during denoising with derivatives
    general_info_dict['n_skip_vols'] = len(np.where(np.sum(np.absolute(parc_ts_dictionary['melodic_mixing_matrix']), axis=1) < 0.1)[0])
    if general_info_dict['n_skip_vols'] == 0:
        general_info_dict['n_skip_vols'] = 1
                
    general_info_dict['mean_fd'] = np.mean(parc_ts_dictionary['confounds']['framewise_displacement'][general_info_dict['n_skip_vols']:])
    general_info_dict['mean_dvars'] = np.mean(parc_ts_dictionary['confounds']['dvars'][general_info_dict['n_skip_vols']:])
    general_info_dict['num_std_dvars_above_1p5'] = len(np.where(parc_ts_dictionary['confounds']['std_dvars'][general_info_dict['n_skip_vols']:] > 1.5)[0])
    general_info_dict['num_std_dvars_above_1p3'] = len(np.where(parc_ts_dictionary['confounds']['std_dvars'][general_info_dict['n_skip_vols']:] > 1.3)[0])
    general_info_dict['num_std_dvars_above_1p2'] = len(np.where(parc_ts_dictionary['confounds']['std_dvars'][general_info_dict['n_skip_vols']:] > 1.2)[0])

    general_info_dict['num_fd_above_0p5_mm'] = len(np.where(parc_ts_dictionary['confounds']['framewise_displacement'][general_info_dict['n_skip_vols']:] > 0.5)[0])
    general_info_dict['num_fd_above_0p4_mm'] = len(np.where(parc_ts_dictionary['confounds']['framewise_displacement'][general_info_dict['n_skip_vols']:] > 0.4)[0])
    general_info_dict['num_fd_above_0p3_mm'] = len(np.where(parc_ts_dictionary['confounds']['framewise_displacement'][general_info_dict['n_skip_vols']:] > 0.3)[0])
    general_info_dict['num_fd_above_0p2_mm'] = len(np.where(parc_ts_dictionary['confounds']['framewise_displacement'][general_info_dict['n_skip_vols']:] > 0.2)[0])
    general_info_dict['labels'] = parc_ts_dictionary['labels']

    
    #Set TR
    general_info_dict['TR'] = parc_ts_dictionary['TR']
        
    #Find session/subject names
    if 'lh_func_path' in parc_ts_dictionary['file_path_dictionary.json'].keys():
    
        temp_path = parc_ts_dictionary['file_path_dictionary.json']['lh_func_path'].split('/')[-1]
        split_end_path = temp_path.split('_')
        
    else:
        
        temp_path = parc_ts_dictionary['file_path_dictionary.json']['nifti_func_path'].split('/')[-1]
        split_end_path = temp_path.split('_')
        
    general_info_dict['subject'] = split_end_path[0]

    if split_end_path[1][0:3] == 'ses':
        general_info_dict['session'] = split_end_path[1]
    else:
        general_info_dict['session'] = []
        
    return general_info_dict
    
    
    
def populate_confounds_dict(file_path_dictionary):
    
        confounds_dictionary = {}
    
        confounds_regressors_tsv_path = file_path_dictionary['confounds_regressors_path']
        confound_df = pd.read_csv(confounds_regressors_tsv_path, sep='\t')
        for (columnName, columnData) in confound_df.iteritems():
            confounds_dictionary[columnName] = columnData.as_matrix()
            
        
        #For convenience, bunch together some commonly used nuisance components
        
        #Six motion realignment paramters
        confounds_dictionary['motion_regs_six'] = np.vstack((confounds_dictionary['trans_x'], confounds_dictionary['trans_y'], confounds_dictionary['trans_z'],
                                         confounds_dictionary['rot_x'], confounds_dictionary['rot_y'], confounds_dictionary['rot_z']))
        
        #Six motion realignment parameters plus their temporal derivatives
        confounds_dictionary['motion_regs_twelve'] = np.vstack((confounds_dictionary['trans_x'], confounds_dictionary['trans_y'], confounds_dictionary['trans_z'],
                                         confounds_dictionary['rot_x'], confounds_dictionary['rot_y'], confounds_dictionary['rot_z'],
                                         confounds_dictionary['trans_x_derivative1'], confounds_dictionary['trans_y_derivative1'],
                                         confounds_dictionary['trans_z_derivative1'], confounds_dictionary['rot_x_derivative1'],
                                         confounds_dictionary['rot_y_derivative1'], confounds_dictionary['rot_z_derivative1']))
        
        #Six motion realignment parameters, their temporal derivatives, and
        #the square of both
        confounds_dictionary['motion_regs_twentyfour'] = np.vstack((confounds_dictionary['trans_x'], confounds_dictionary['trans_y'], confounds_dictionary['trans_z'],
                                         confounds_dictionary['rot_x'], confounds_dictionary['rot_y'], confounds_dictionary['rot_z'],
                                         confounds_dictionary['trans_x_derivative1'], confounds_dictionary['trans_y_derivative1'],
                                         confounds_dictionary['trans_z_derivative1'], confounds_dictionary['rot_x_derivative1'],
                                         confounds_dictionary['rot_y_derivative1'], confounds_dictionary['rot_z_derivative1'],
                                         confounds_dictionary['trans_x_power2'], confounds_dictionary['trans_y_power2'], confounds_dictionary['trans_z_power2'],
                                         confounds_dictionary['rot_x_power2'], confounds_dictionary['rot_y_power2'], confounds_dictionary['rot_z_power2'],
                                         confounds_dictionary['trans_x_derivative1_power2'], confounds_dictionary['trans_y_derivative1_power2'],
                                         confounds_dictionary['trans_z_derivative1_power2'], confounds_dictionary['rot_x_derivative1_power2'],
                                         confounds_dictionary['rot_y_derivative1_power2'], confounds_dictionary['rot_z_derivative1_power2']))
        
        #white matter, and csf
        confounds_dictionary['wmcsf'] = np.vstack((confounds_dictionary['white_matter'], confounds_dictionary['csf']))
        
        #white matter, csf, and their temporal derivatives
        confounds_dictionary['wmcsf_derivs'] = np.vstack((confounds_dictionary['white_matter'], confounds_dictionary['csf'], 
                                      confounds_dictionary['white_matter_derivative1'], confounds_dictionary['csf_derivative1']))
        
        #White matter, csf, and global signal
        confounds_dictionary['wmcsfgsr'] = np.vstack((confounds_dictionary['white_matter'], confounds_dictionary['csf'], confounds_dictionary['global_signal']))
        
        #White matter, csf, and global signal plus their temporal derivatives
        confounds_dictionary['wmcsfgsr_derivs'] = np.vstack((confounds_dictionary['white_matter'], confounds_dictionary['csf'], confounds_dictionary['global_signal'],
                                        confounds_dictionary['white_matter_derivative1'], confounds_dictionary['csf_derivative1'],
                                        confounds_dictionary['global_signal_derivative1']))
        
        #The first five anatomical comp cor components
        confounds_dictionary['five_acompcors'] = np.vstack((confounds_dictionary['a_comp_cor_00'], confounds_dictionary['a_comp_cor_01'],
                                         confounds_dictionary['a_comp_cor_02'], confounds_dictionary['a_comp_cor_03'],
                                         confounds_dictionary['a_comp_cor_04']))
        
        return confounds_dictionary
    
    
    
def save_dictionary(dictionary, path_for_dictionary_dir, overwrite = False):
    
        
    if path_for_dictionary_dir[-5:] == '.json':
        
        with open(os.path.join(path_for_dictionary_dir), 'w') as temp_file:
            json_dict = json.dumps(dictionary, indent = 4, sort_keys = True)
            temp_file.write(json_dict)        

    else:
        
        if os.path.exists(path_for_dictionary_dir) == True:
        
            if overwrite == False:

                raise NameError('Error: Parc Timeseries Dictionary Directory Already Exists at this path')

        else:
        
            os.makedirs(path_for_dictionary_dir)
        
        
        
        for temp_key in dictionary.keys():


            if type(dictionary[temp_key]) == dict:

                to_overwrite = overwrite
                save_dictionary(dictionary[temp_key], os.path.join(path_for_dictionary_dir, temp_key), overwrite = to_overwrite)

            else:


                if path_for_dictionary_dir[-3:] == 'txt':

                    with open(os.path.join(path_for_dictionary_dir, temp_key), 'w') as temp_file:
                        temp_file.write(str(dictionary[temp_key]))

                else:

                    np.save(os.path.join(path_for_dictionary_dir, temp_key), dictionary[temp_key])
                 
            
    return
                
                
    
    
    
def load_dictionary(dictionary_dir_path):
    
    dictionary = {}
    
    os.chdir(dictionary_dir_path)
    directory_contents = os.listdir()
    directory_final_name = dictionary_dir_path.split('/')[-1]
    for temp_file in directory_contents:
        
        if os.path.isdir(temp_file):
            
            dictionary[temp_file] = load_dictionary(os.path.join(dictionary_dir_path, temp_file))
            os.chdir(dictionary_dir_path)
            
        else:
            
            if temp_file[-5:] == '.json':
                                
                with open(temp_file, 'r') as temp_json:
                    json_contents = temp_json.read()
                    dictionary[temp_file] = json.loads(json_contents)
                    
            else:
            
                if dictionary_dir_path[-3:] == 'txt':

                    with open(temp_file,'r') as temp_reading:
                        file_contents = temp_reading.read()

                        try:
                            dictionary[temp_file.split('.')[0]] = float(file_contents)
                        except:

                            if file_contents[0] == '[' and file_contents[-1] == ']':
                                split_contents = file_contents[1:-1].split(',')
                                dictionary[temp_file.split('.')[0]] = []
                                for temp_item in split_contents:
                                    split_limited_1 = temp_item.replace(' ','')
                                    split_limited_2 = split_limited_1.replace("'","")
                                    dictionary[temp_file.split('.')[0]].append(split_limited_2)

                            else:
                                dictionary[temp_file.split('.')[0]] = file_contents                

                else:

                    dictionary[temp_file.split('.')[0]] = np.load(temp_file)
                
                
                
    return dictionary
